// compproc.cc

#include "compproc.h" 
 
void orderreal(bigfloat& e1, bigfloat& e2, bigfloat& e3)  // puts in decreasing order
{ 
  bigfloat t;
  if      (e1 < e3) {t=e1; e1=e3; e3=t;}      //swap(e1,e3);
  if      (e1 < e2) {t=e1; e1=e2; e2=t;}      //swap(e1,e2);
  else if (e2 < e3) {t=e2; e2=e3; e3=t;}      //swap(e2,e3);
}
 
bigcomplex cagm(const bigcomplex& a, const bigcomplex& b)
{
     bigcomplex x=a, y=b, z;
     bigfloat theta, two(2), pi=Pi();
     while (!is_approx_zero(abs(x-y)))
      { z=(x+y)/two;
        y= sqrt(x*y);
        theta = two*arg(y/z);
        if (theta>pi|| theta<=-pi) y=-y;   
        x=z;
//	cout<<x<<endl;
      }
   return x;
}

bigcomplex normalize(bigcomplex& w1, bigcomplex& w2)
{
   bigcomplex tau = w1/w2, w3;
   if (tau.imag() < 0) { w1=-w1 ; tau=-tau ; }
   w1=w1-w2*round(tau.real());
   tau=w1/w2;
 
   for (int i=1;i<50 && (abs(tau)<1);i++)  
// {Just to stop infinite loop due to rounding}
   { w3=-w1; w1=w2; w2=w3; tau=w1/w2;
    w1=w1-w2*round(tau.real());
    tau=w1/w2;
   }
   return tau;
}
 
void getc4c6(const bigcomplex& w1, const bigcomplex& w2, 
	     bigcomplex& c4, bigcomplex &c6)
{
  bigcomplex tau= w1/w2;
  bigfloat zero(0), one(1), two(2);
  bigfloat pi(Pi());
  bigfloat x = two*pi*tau.real();
  bigfloat y = two*pi*tau.imag();
  bigcomplex q = exp(-y) *  bigcomplex(cos(x),sin(x));
  bigcomplex f = two*pi/w2;
  bigcomplex f2 = f*f;  
  bigcomplex f4=f2*f2;
  
  bigcomplex term   = bigcomplex(one); 
  bigcomplex qpower = bigcomplex(one);
  bigcomplex sum4   = bigcomplex(zero);
  bigcomplex sum6   = bigcomplex(zero);
  
  bigfloat n, n2;

  for (n=1; 
#ifdef LiDIA
       !is_approx_zero(term); 
#else
       !is_zero(term); 
#endif
       n+=1)
    {  n2      = n*n;
       qpower *= q;
       term    = n*n2*qpower/(one-qpower);
       sum4   += term;
       term   *= n2;
       sum6   += term;
     }
  c4= (one +  bigfloat(240)*sum4)*f4;
  c6= (one -  bigfloat(504)*sum6)*f4*f2;
}
 
bigcomplex discriminant(const bigcomplex& b, const bigcomplex& c, const bigcomplex& d)
{
    bigcomplex bb = b*b, cc = c*c, bc = b*c;
    return bigfloat(27)*d*d - bc*bc +  bigfloat(4)*bb*b*d -  bigfloat(18)*bc*d +  bigfloat(4)*c*cc;
}
 
//#define DEBUG_CUBIC

bigcomplex* solvecubic(const bigcomplex& c1, const bigcomplex& c2, const bigcomplex& c3) //Returns an array
{                                                       //of 3 roots.
#ifdef DEBUG_CUBIC
cout << "In solvecubic with c1 = "<<c1<<", c2 = "<<c2<<", c3 = "<<c3<<"\n";
#endif
  bigfloat three(3), two(2), one(1);
  bigfloat third = one/three;
   bigcomplex w =  bigcomplex(-1, sqrt(three))/two;
   bigcomplex disc = discriminant(c1,c2,c3);
   bigcomplex p3=  three*c2 - c1*c1;
   bigcomplex mc1over3 = -c1*third;
#ifdef DEBUG_CUBIC
cout << "p3 = "<<p3<<", -c1/3 = "<<mc1over3<<"\n";
#endif
   bigcomplex *roots = new bigcomplex[3];
 
   if (is_zero(abs(disc))) 
     {if (is_zero(abs(p3)))     // triple root
        {roots[0]=roots[1]=roots[2]= mc1over3;
        }
      else           // double root
        {roots[0]=roots[1]= (c1*c2 -  bigfloat(9)*c3)/( p3+p3);
         roots[2]=-( roots[0] + roots[0]+c1);
        }
     }
   else              // distinct roots
     {
      bigcomplex q = (((mc1over3+c1)*mc1over3 +c2)*mc1over3 +c3); 
                    // = F(mc1over3);
      if (is_approx_zero(abs(p3)))       // pure cubic
        { 
#ifdef DEBUG_CUBIC
cout << "In pure cubic case\n";
cout << "About to take cube root of -q = " << (-q) << " by pow(-,third) where third = " << third << endl;
#endif
	  roots[0]=pow(-q,third);
#ifdef DEBUG_CUBIC
cout << "returns result " << roots[0] << endl;
#endif
          roots[1]=w*roots[0];
          roots[2]=w*roots[1];
          roots[0]+=mc1over3;
          roots[1]+=mc1over3;
          roots[2]+=mc1over3;
        }
      else
        {bigcomplex d =  bigfloat(729)*q*q+ bigfloat(4)*p3*p3*p3;
#ifdef DEBUG_CUBIC
cout << "q = " << q << ", p3 = " << p3 << ", d = " << d << "\n";
#endif
         bigcomplex t1cubed = bigfloat(0.5)*(sqrt(d)- bigfloat(27)*q);
#ifdef DEBUG_CUBIC
cout << "About to take cube root of " << t1cubed << " by power(-,third) where third = " << third << endl;
#endif
         bigcomplex t1 = pow(t1cubed,third);
#ifdef DEBUG_CUBIC
cout << "returns result " << t1 << endl;
#endif
         bigcomplex t2 = t1*w;
         bigcomplex t3 = t2*w;
#ifdef DEBUG_CUBIC
cout<<"t1cubed = "<<t1cubed<<"\n";
cout<<"t1 = "<<t1<<", t2 = "<<t2<<", t3 = "<<t3<<"\n";
#endif
         roots[0] = (-c1+t1-p3/t1)* third;
         roots[1] = (-c1+t2-p3/t2)* third;
         roots[2] = (-c1+t3-p3/t3)* third;
        }
    }
  long i, iter, niter=2; // number of iterations in Newton refinement of roots
#ifdef DEBUG_CUBIC
cout << "refining roots using Newton with " << niter << "iterations\n";
cout << "unrefined roots: ";
for(i=0; i<3;i++) cout << roots[i] << "\n";
#endif
for(i=0; i<3; i++)
  {
    bigcomplex z = roots[i], fz, fdashz;
    for(iter=0; iter<niter; iter++)
      {
	fz = ((z+c1)*z+c2)*z+c3;
	fdashz = (3*z+2*c1)*z+c2;
	if(!is_zero(fdashz)) z -= fz/fdashz;
      }
    roots[i] = z;
  }
#ifdef DEBUG_CUBIC
cout << "refined roots: ";
for(i=0; i<3;i++) cout << roots[i] << "\n";
#endif

  return roots;
}
 
bigcomplex* solverealquartic(const bigfloat& a, const bigfloat& b, const bigfloat& c, const bigfloat& d, const bigfloat& e)
{
#ifdef DEBUG
  cout<<"In solverealquartic with (a,b,c,d,e)=("<<a<<","<<b<<","<<c<<","<<d<<","<<e<<")\n";
#endif
  bigfloat ii = 12*a*e - 3*b*d + c*c;
  bigfloat jj =  (72*a*e + 9*b*d - 2*c*c) * c - 27*(a*d*d + b*b*e);
#ifdef DEBUG
  cout<<"ii="<<ii<<"\njj="<<jj<<"\n";
#endif
  bigfloat disc = 4*ii*ii*ii-jj*jj;
  bigfloat  H = 8*a*c - 3*b*b, R = b*b*b + 8*d*a*a - 4*a*b*c;
  bigfloat  Q = H*H-16*a*a*ii;  // = 3*Q really
  int type, nrr;
  if(disc<0) 
    {type=3; nrr=2;}       // 2 real roots
  else 
    {
      if((H<0)&&(Q>0)) 
	{type=2; nrr=4;}   // 4 real roots
      else 
	{type=1; nrr=0;}   // 0 real roots
    }
  bigcomplex c1(0), c2(-3*ii), c3(jj);
#ifdef DEBUG
  cout<<"Type = " << type << " ("<<nrr<<" real roots)\n";
  cout<<"Coeffs of resolvent cubic are:\n"<<c1<<"\n"<<c2<<"\n"<<c3<<endl;
#endif
  bigcomplex* cphi = solvecubic( c1, c2, c3);
  bigcomplex* roots = new bigcomplex[4];
  bigfloat a4=4*a;
  bigfloat oneover4a = bigfloat(1)/a4;
#ifdef DEBUG
  cout<<"Roots of cubic are:\n"<<cphi[0]<<"\n"<<cphi[1]<<"\n"<<cphi[2]<<endl;
#endif
  
  if(type<3) 
    {
#ifdef DEBUG
      cout<<"Positive discriminant\n";
#endif
      // all the phi are real;  order them so that a*phi[i] decreases
      bigfloat phi1 = real(cphi[0]);
      bigfloat phi2 = real(cphi[1]);
      bigfloat phi3 = real(cphi[2]);
      if(a>0)      orderreal(phi1,phi2,phi3); 
      else         orderreal(phi3,phi2,phi1); 
#ifdef DEBUG
      cout<<"phi = "<<phi1<<", "<<phi2<<", "<<phi3<<"\n";
#endif

      if(type==2) // all roots are real
	{
#ifdef DEBUG
	  cout<<"Type 2\n";
#endif
	  bigfloat r1 = sqrt((a4*phi1-H)/bigfloat(3));
	  bigfloat r2 = sqrt((a4*phi2-H)/bigfloat(3));
	  bigfloat r3 = sqrt((a4*phi3-H)/bigfloat(3));
	  if(R<0)  r3 = -r3;
#ifdef DEBUG
	  cout<<"r_i = "<<r1<<", "<<r2<<", "<<r3<<"\n";
	  cout<<"product = "<<r1*r2*r3<<", R = "<<R<<endl;
#endif
	  roots[0] = bigcomplex(( r1 + r2 - r3 - b) * oneover4a);
	  roots[1] = bigcomplex(( r1 - r2 + r3 - b) * oneover4a);
	  roots[2] = bigcomplex((-r1 + r2 + r3 - b) * oneover4a);
	  roots[3] = bigcomplex((-r1 - r2 - r3 - b) * oneover4a);
	  // Those are all real and in descending order of size
	}
      else // no roots are real
	{
#ifdef DEBUG
	  cout<<"Type 1\n";
#endif
	  bigfloat r1 = sqrt((a4*phi1-H)/3);
	  bigfloat ir2 = sqrt(-((a4*phi2-H)/3));
	  bigfloat ir3 = sqrt(-((a4*phi3-H)/3));
	  if(R>0)  r1 = -r1;
#ifdef DEBUG
	  cout<<"r_i = "<<r1<<", "<<ir2<<"i, "<<ir3<<"i\n";
	  cout<<"product = "<<-r1*ir2*ir3<<", R = "<<R<<endl;
#endif
	  roots[0] = bigcomplex( r1-b,  ir2 - ir3) * oneover4a;
	  roots[1] = conj(roots[0]); // bigcomplex( r1-b,  ir2 - ir2) * oneover4a;
	  roots[2] = bigcomplex(-r1-b,  ir2 + ir3 ) * oneover4a;
	  roots[3] = conj(roots[2]); // bigcomplex(-r1-b, -ir2 - ir3 ) * oneover4a;
	}
    }
  else // disc < 0
    {
#ifdef DEBUG
      cout<<"Negative discriminant\nType 3\n";
#endif
      bigfloat realphi;       // will hold the real root, which will be cphi[2]
      if (is_real(cphi[1])) 
	{
	  realphi=real(cphi[1]);
	  cphi[1]=cphi[2];
	  cphi[2]=realphi;
	}
      else 
	if (is_real(cphi[2])) 
	  {
	    realphi=real(cphi[2]);
	  }
	else 
	  {
	    realphi=real(cphi[0]);
	    cphi[0]=cphi[2];
	    cphi[2]=realphi;
	  }
#ifdef DEBUG
      cout<<"Sorted roots of cubic (real one last) are \n";
      cout<<cphi[0]<<"\n"<<cphi[1]<<"\n"<<cphi[2]<<endl;
#endif
      bigcomplex r1 = sqrt((a4*cphi[0]-H)/3);
      bigfloat r3   = sqrt((a4*realphi-H)/3);
      if(R<0)  r3 = -r3;
#ifdef DEBUG
      cout<<"r_i = "<<r1<<", "<<conj(r1)<<", "<<r3<<"\n";
      cout<<"product = "<<r1*conj(r1)*r3<<", R = "<<R<<endl;
#endif
      roots[0] = bigcomplex( r3 - b, 2*imag(r1) ) * oneover4a;
      roots[1] = conj(roots[0]);
      roots[2] = bigcomplex(( 2*real(r1) - r3 - b)) * oneover4a;
      roots[3] = bigcomplex((-2*real(r1) - r3 - b)) * oneover4a;
      // roots[2] and roots[3] are real
    }
  delete[] cphi;
  return roots;
}


bigcomplex* solvequartic(const bigcomplex& a, const bigcomplex& b, const bigcomplex& c, const bigcomplex& d)
{ bigcomplex p,q,r,aa,e,f1,f2;
  bigcomplex a4=a/4;
  bigcomplex* roots = new bigcomplex[4];
  if(is_zero(d))
    { 
      roots[0]= 0;
      bigcomplex* cuberoots=solvecubic(a,b,c);
      roots[1] = cuberoots[0];
      roots[2] = cuberoots[1];
      roots[3] = cuberoots[2];
      delete[] cuberoots;
    }
    else
      {
	p = b -  3*a*a / 8;
	q = ((a / 2) * (a*a4 - b)) + c;
	r = ((256*d) - (64*a*c) + (16*a*a*b) - (3*a*a*a*a)) / 256;
	if( is_approx_zero(q) ) 
	  {
	    bigcomplex s = sqrt(p*p - 4*r);
	    roots[0] =  sqrt((-p + s) / 2) - a4;
	    roots[1] = -sqrt((-p + s) / 2) - a4;
	    roots[2] =  sqrt((-p - s) / 2) - a4;
	    roots[3] = -sqrt((-p - s) / 2) - a4;
	  }
      else
	{
	  bigcomplex* aaroots= solvecubic(-p / 2,-r,((p * r) / 2 - (q * q) / 8));
	  aa = aaroots[0];
	  delete[] aaroots;
	  if( is_approx_zero(aa) ) aa=0;
	  e = sqrt(-p + 2*aa);
	  f1 = (aa + q / (2*e));
	  f2 = (aa - q / (2*e));
	  bigcomplex s1 = sqrt(e*e -  4*f1);
	  bigcomplex s2 = sqrt(e*e -  4*f2);
	  roots[0] = (( e + s1) /  2) - a4;
	  roots[1] = (( e - s1) /  2) - a4;
	  roots[2] = ((-e + s2) /  2) - a4;
	  roots[3] = ((-e - s2) /  2) - a4;
	}
      }

  long i, iter, niter=2; // number of iterations in Newton refinement of roots
#ifdef DEBUG_QUARTIC
  cout << "refining roots using Newton with " << niter << "iterations\n";
  cout << "unrefined roots: ";
  for(i=0; i<4;i++) cout << roots[i] << "\n";
#endif
  for(i=0; i<4; i++)
    {
      bigcomplex z = roots[i], fz, fdashz;
      for(iter=0; iter<niter; iter++)
	{
	  fz = (((z+a)*z+b)*z+c)*z+d;
	  fdashz = ((4*z+3*a)*z+2*b)*z+c;
	  if(!is_zero(fdashz)) z -= fz/fdashz;
	}
      roots[i] = z;
    }
#ifdef DEBUG_QUARTIC
  cout << "refined roots: ";
  for(i=0; i<4;i++) cout << roots[i] << "\n";
#endif
  return roots;
}

void quadsolve(const bigfloat& p, const bigfloat& q, 
	       bigcomplex& root1, bigcomplex& root2)
{
  bigcomplex disc(p*p- bigfloat(4)*q);
  bigcomplex rootdisc = sqrt(disc);
  root1 = ( rootdisc-p)/ bigfloat(2);
  root2 = (-rootdisc-p)/ bigfloat(2);
}
 
int norepeats(long* x, int n)
{long* ans = new long[n];
 int nr, newone, i, j;
 long xi;
 nr = 0;
 for (i=0; i<n; i++)
 {xi = x[i];  newone = 1;
  for (j=0; (j<i) && newone; j++) newone = (xi!=x[j]);
  if (newone) ans[nr++]=xi;
 }
 for (i=0; i<nr; i++) x[i]=ans[i];
 delete ans;
 return nr;
}

long* introotscubic(long a, long b, long c, int& nr)
{ bigcomplex za(a), zb(b), zc(c);
  bigcomplex* croots =  solvecubic(za,zb,zc);
  long* iroots=new long[3];
  int i,niroots =0; long x,cx;
  for (i=0; i<3; i++)
    {
      cout << "Complex root = " << croots[i] << endl;
      bigfloat xx = croots[i].real();
      Iasb(x,xx);
      cout << "Rounds to " << x << endl;
    if (x==0) {if (c==0) iroots[niroots++]=0;}
    else
        {cx = c/x;
         if (x*cx==c)
            if (((x+a)*x+b+cx) ==  0)
               iroots[niroots++]=x;
        }
     }
  nr = norepeats(iroots,niroots);
  delete[] croots;
  return iroots;  
}

