//
// isogs.h -- header for mapping points from lattices, and isogeny routines
//

#ifndef _ISOGS_H_
#define _ISOGS_H_

#include "cperiods.h"  // which itself includes everything relevant

// isogeny functions:

CurveRedArray lisog(const CurveRed& CR, Cperiods& cp,
			  long ell, int verbose=0);
			  // cp must be normalized for lattice at call time
			  // returned array is of length up to 3
			  // of curves ell-isogenous to CR


int semistable(const CurveRed& CR);

longlist getelllist(const CurveRed& CR); 
// returns list of possible primes l ("ell") for which CR might have an l-isogeny.

#define MAXNCURVES 26    // max number of curves in an isogeny class.

class IsogenyClass {
private: 
  CurveRedArray curves;
  longlist llist; 
  long nell, ncurves, ndone;
  int ss, verb;
  Cperiods cp;
  longlist fromlist; // fromlist[i]=j if curve i first constructed from 
  longlist isoglist; // curve j via an isoglist[i]-isogeny
  longlist matij;
  void matset(long i, long j, long ell) { matij[i*MAXNCURVES+j]=ell;}
  long mat(long i, long j)const {return matij(i*MAXNCURVES+j);}
  void process(long i);  // process i'th curve
public:
  IsogenyClass(const CurveRed& C, int verbose);
  void grow(void);       // does the work
  void display_llist(ostream& os)const {os<<llist;}
  void displaycurves(ostream& os)const;
  void displaymat(ostream& os)const;
  void display(ostream& os)const {displaycurves(cout); displaymat(cout);}
  void dumpdata(ostream& os, long rank);  // output for textab to input
  longlist getmat() const;
  CurveRedArray getcurves() const {return curves;}
};


#endif
