/* M1RANK.CC:  main program (Integer version); calls mrank1 always  */
/*                                                   */
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <builtin.h>
#include "points.h"  // from qcurves library
#include "mquartic.h"
#include "mrank1.h"

Integer a1,a2,a3,a4,a6;
Curve C;
Curvedata CD;
int verbose, traceabc, traceequiv, ptl, sha2, sha2dash, rank, success, naux;
// Increase the following if you need to search further for rational
// points on quartics (logarithmic!)
long hlimq=10; 

int getcurve(void)
{
  if (verbose) cout << "Enter curve coefficients a1,a2,a3,a4,a6 ?" << endl;
  cin >> a1 >> a2 >> a3 >> a4 >> a6;
  C = Curve(a1,a2,a3,a4,a6);
  CD = Curvedata(C,1);     // "1" means minimise
  return (a1!=0||a2!=0||a3!=0||a4!=0||a6!=0);
}

int main()
{
  cout << "Program m1rank: uses 2-descent (NOT via 2-isogeny even if possible) to\n";
  cout << "determine the rank of an elliptic curve E over Q, and list a\n";
  cout << "set of points which generate E(Q) modulo 2E(Q).\n";
  cout << "\nFor details of algorithms see the author's book.\n";
  cout << "\nPlease acknowlege use of this program in published work, \n";
  cout << "and send problems to cremona@maths.exeter.ac.uk.\n\n";

  initprimes(100000);
  struct tms buffer;
  long starttime,stoptime;
  verbose=1; traceabc=0; traceequiv=0; ptl=1; naux=5;
  cout << "Verbose mode? (0/1)\n"; cin >> verbose;
  cout << "Bound on height in quartic point search (e.g. 5)? "; cin >> hlimq;
//  cout << "Number of primes to use for (a,b,c) sieving? "; cin>>naux;
//  cout << "List points? (0/1)\n"; cin >> ptl;
//  cout << "Trace abc testing? (0/1)\n"; cin >> traceabc;
//  cout << "Trace equivalence testing? (0/1)\n"; cin >> traceequiv;
  cin.flags( cin.flags() | ios::dec );  //force decimal input (bug fix)
  
  while (getcurve())
    {
      if (verbose) cout << endl;
      if(C!=(Curve)CD)
	{
	  cout<<"Input curve "<<C<<"\n";
	  cout<<"Working with minimal curve "<<(Curve)CD<<"\n";
	}
      else 
	cout << "Curve "<< C <<" :  ";
      if (verbose) cout << endl;
      times(&buffer);
      starttime = buffer.tms_utime;

 
      rank1 r1(&CD, verbose, traceabc, traceequiv, 20, hlimq, naux);
      
      if (r1.ok())
	{    
	  cout << "Rank = " << r1.getrank() << "\t";
	  if (sha2>1) cout << "|Sha_E[2]| = " << r1.getsha() << "\t";
	  if(ptl) 
	    {
	      cout<<"\n"; 
	      r1.listpoints();
//	      PointArray plist = r1.getpoints();
//	      cout << "\n\nDuplicate list of points: \n" << plist <<"\n";
	    }
	}
      else cout << "Failed to compute rank\n";
      times(&buffer);
      stoptime = buffer.tms_utime;
      cout << "\t(cpu time = " << ((stoptime-starttime)*1./HZ) << " seconds)" << endl;
    }
}
