#include "curve.h"
#include "silver.h"

double logplus(double x)
{
  double ax = abs(x);
  if(ax<1) return 0;
  return log(ax);
}

double hj(const Curvedata& CD, double& realjay)
{
  bigint c4, c6, njay, djay;
  c4=getc4(CD);
  c6=getc6(CD);
  njay = pow(c4,3);
  djay = getdiscr(CD);
  if((djay==0)||(njay==0)) {realjay=0; return 0;}

  double g = I2double(gcd(njay,djay));
  double xnjay = I2double(njay)/g;
  double xdjay = I2double(djay)/g;

  realjay = xnjay/xdjay;

  double x = log(abs(xnjay));
  double y = log(abs(xdjay));

  if(x<y) return y;
  else    return x;
}

double silver(const Curvedata& CD)
{
  bigint b2 = getb2(CD);
  bigint delta = getdiscr(CD);
  double realjay;
  double hjay = hj(CD,realjay);

// NB the constant 1.922 = 2*0.961 below is from Bremner's correction 
// to Silverman's paper; Silverman has 0.973 givin 2*0.973 = 1.946.

  double mu = 1.922  + hjay/12
                     + log(abs(I2double(delta)))/6
                     + logplus(realjay)/6
	             + logplus(I2double(b2)/12);
  
  if(b2!=0) mu += log(2.0);

  return mu;
}


// Only silverman done at present
double height_constant(const Curvedata& CD)
{
  return silver(CD);
// Will actually return minimum of silverman and samir
}
