#include "marith.h"
#include "sqfdiv.h"

//#define DEBUG

sqfdiv::sqfdiv(const bigint& dd, int posd, bigintArray* plist)
:positive(posd), primebase(plist), d(1), factor(0), np(0)
{
  bigint p;
  for(long i=0; i<plist->length; i++)
    if(p=(*primebase)[i],div(p,dd)) {d*=p; np++;}
  maxnsub=2<<np;    // = 2^(np+1) 
  maxngens=np+1; ngens=0;  nsub=1;
  subgp = bigintArray(maxnsub);
  gens  = bigintArray(maxngens);
  pivs = longlist(maxngens);
  subgp[0]=1;
  if(positive) {nsub=2; subgp[1]=-1; ngens=1; gens[0]=-1; pivs[0]=-1;}
}

void sqfdiv::usediv(const bigint& ee)
{
  bigint e = sqfred(ee,*primebase);
#ifdef DEBUG
  cout << "usediv called with e = " << ee 
    << " reduced mod squares to "<<e<<"\n";
#endif
  int triv=(e==1); long i;
  for(i=0; (i<ngens)&&!triv; i++)
    {
      long pivi = pivs[i];
      if(pivi==-1) {e=abs(e);}
      else { const bigint& p = (*primebase)[pivi];
	     if(div(p,e)) e = sqfmul(e,gens[i]);
	   }
      triv = (e==1);
#ifdef DEBUG
      cout << "After using gen "<<gens[i]<<", reduced to " << e <<"\n";
#endif
    }
  if(triv) 
    {
#ifdef DEBUG
      cout << ee << " is in subgroup already...\n";
#endif
      return;
    }
#ifdef DEBUG
  cout << ee << " not in subgroup; reduced to " << e << "; adding...\n";
#endif
// update subgroup and gens:
  gens[ngens] = e;
  for(i=0; i<nsub; i++)
    {
      subgp[i+nsub]=sqfmul(subgp[i],e);
    }
  nsub*=2;

// find new pivotal prime:

  bigint p; long valp=0, newpiv;;
  for(i=primebase->length; (i>0)&&(!valp); i--)
    {
      p=(*primebase)[i-1];
      if(div(p,d)) {valp=val(p,e); newpiv=i-1;}
    }
  // now if valp!=0, p|d and p|e to an odd power
  if(valp) 
    {d/=p; np--; factor++; pivs[ngens++]=newpiv;}
  else 
    if((e<0)&& !positive) 
      {positive=1; factor++; pivs[ngens++]=-1;}
  //else e is square over the support of d so should have returned earlier!
#ifdef DEBUG
  cout << "New gens:     " << bigintArray(gens,ngens) << endl;
  cout << "New pivs:     " << longlist(pivs,ngens) << endl;
  cout << "New subgroup: " << bigintArray(subgp,nsub) << endl;
#endif
}

bigintArray sqfdiv::getdivs() const
{
 long nd=1<<np;
 if(!positive) nd*=2;
// cout << "Constructing divisor list for d = " << d;
// cout << ", positive = " << positive;
// cout << ", number of divisors = " << nd << endl;
 bigintArray dlist(nd);
 dlist[0]=1; 
 nd=1;
 if(!positive) {dlist[nd++]=-1;}
 for(long i=0; i<primebase->length; i++)
   {
     const bigint& p = (*primebase)[i];
     if(ndiv(p,d)) continue;
     for (long k=0; k<nd; k++)
       dlist[nd+k] = p*dlist[k];
     nd*=2;
   }
 return dlist;
}

bigintArray sqfdiv::getsupp(int bothsigns) const
{
 long ns=np;
 int use_minus_one = (!positive)||bothsigns;
 if(use_minus_one) ns++;
 bigintArray supp(ns);
 ns=0;
 if(use_minus_one) {supp[ns++]=-1;}
 for(long i=0; i<primebase->length; i++)
   {
     const bigint& p = (*primebase)[i];
     if(ndiv(p,d)) continue;
     supp[ns++] = p;
   }
 return supp;
}

void sqfdiv::display()
{
  cout << "Current reduced d = " << d << "\n";
  cout << "np = " << np << ", positive = " << positive << ", log_2(factor) = ";
  cout << factor << "\n";
  cout << "Subgroup gens     = " << bigintArray(gens,ngens) << endl;
  cout << "Subgroup elements = " << bigintArray(subgp,nsub) << endl;
}

bigint sqfred(const bigint& a, const bigintArray& plist)
{
  bigint ans(sign(a));
  for(long i=0; i<plist.length; i++)
    {
     const bigint& p = plist(i);
     if(odd(val(p,a))) ans*=p;
    }
  return ans;
}

bigint sqfmul(const bigint& a, const bigint& b)
{ const bigint& g = gcd(a,b);
  const bigint& ans = (a/g)*(b/g);
  return ans;
}

bigint makenum(const bigintArray& supp, long mask)
{
  bigint ans(1); long i, ns=supp.length;
  for(i=0; i<ns; i++) if(testbit(mask,i)) ans=sqfmul(ans,supp(i));
  return ans;
}

long makeindex(const bigintArray& supp, const bigint& n, bigint& n0)
{
  long i, ns = supp.length, index=0;
  n0=1;
  if(is_zero(n)) return index;
  for(i=0; i<ns; i++)
    {
      bigint pi = supp(i);
      if(sign(pi)<0) // special case, supp might have -1 as well as primes
	{
	  if(sign(n)<0) {setbit(index,i); n0=-n0;}
	}
      else // pi is really a prime
	{
	  if(odd(val(pi,n))) {setbit(index,i); n0*=pi;}
	}
    }
  return index;
}

// support(n) is like pdivs(n) but includes -1 always (except n=0, 
//but it should never be called with 0)
bigintArray support(const bigint& n)
{
  if(is_zero(n)) {bigintArray supp(0); return supp;}
  bigintArray supp_pos  = pdivs(n);
  long ns=supp_pos.length;
  bigintArray supp(ns+1);  supp[0]=-1;
  while(ns--) supp[ns+1]=supp_pos[ns];
  return supp;
}
